@@ -37,8 +37,8 @@ class BuildConfigurationInformation(object):
     """
     version_major = 1
     version_minor = 8
-    version_patch = 11
-    version_so_patch = 2
+    version_patch = 14
+    version_so_patch = 13
     version_suffix = ''
 
     version_string = '%d.%d.%d%s' % (
@@ -58,7 +58,7 @@ class BuildConfigurationInformation(object):
         self.include_dir = os.path.join(self.build_dir, 'include')
         self.full_include_dir = os.path.join(self.include_dir, 'botan')
 
-        all_files = sum([mod.add for mod in modules], [])
+        all_files = sum([list(mod.add) for mod in modules], [])
 
         self.headers = sorted(
             [file for file in all_files if file.endswith('.h')])
@@ -161,6 +161,10 @@ def process_command_line(args):
                            action='store_true', default=False,
                            help=SUPPRESS_HELP)
 
+    build_group.add_option('--link-method',
+                           default=None,
+                           help=SUPPRESS_HELP)
+
     mods_group = OptionGroup(parser, 'Module selection')
 
     mods_group.add_option('--enable-modules', dest='enabled_modules',
@@ -286,7 +290,7 @@ def lex_me_harder(infofile, to_obj, allowed_groups, name_val_pairs):
 
     for group in allowed_groups:
         to_obj.__dict__[group] = []
-    for (key,val) in name_val_pairs.iteritems():
+    for (key,val) in name_val_pairs.items():
         to_obj.__dict__[key] = val
 
     def lexed_tokens(): # Convert to an interator
@@ -409,8 +413,8 @@ class ArchInfo(object):
             self.unaligned_ok = 0
 
     def all_submodels(self):
-        return sorted(zip(self.submodels, self.submodels) +
-                          self.submodel_aliases.items(),
+        return sorted([(k,k) for k in self.submodels] +
+                      [k for k in self.submodel_aliases.items()],
                       key = lambda k: len(k[0]), reverse = True)
 
     def defines(self, target_submodel, with_endian):
@@ -509,6 +513,18 @@ class CompilerInfo(object):
 
         return ''
 
+    def library_opt_flags(self, options):
+        def gen_flags():
+            yield self.lib_opt_flags
+
+            if options.debug_build:
+                yield self.debug_flags
+
+            if not options.debug_build:
+                yield self.no_debug_flags
+
+        return (' '.join(gen_flags())).strip()
+
     def so_link_command_for(self, osname):
         if osname in self.so_link_flags:
             return self.so_link_flags[osname]
@@ -558,13 +574,26 @@ class OsInfo(object):
                 for feat in self.target_features]
 
 def canon_processor(archinfo, proc):
+    # First, try to search for an exact match
     for ainfo in archinfo.values():
         if ainfo.basename == proc or proc in ainfo.aliases:
             return (ainfo.basename, ainfo.basename)
-        else:
-            for (match,submodel) in ainfo.all_submodels():
-                if re.search(match, proc) != None:
-                    return (ainfo.basename, submodel)
+
+        for (match,submodel) in ainfo.all_submodels():
+            if proc == submodel:
+                return (ainfo.basename, submodel)
+
+    # Now, try searching via regex match
+    for ainfo in archinfo.values():
+        for (match,submodel) in ainfo.all_submodels():
+            if re.search(match, proc) != None:
+                return (ainfo.basename, submodel)
+
+    logging.debug('Known CPU names: ' + ' '.join(
+        sorted(sum([[ainfo.basename] + \
+                    ainfo.aliases + \
+                    [x for (x,_) in ainfo.all_submodels()]
+                    for ainfo in archinfo.values()], []))))
 
     raise Exception('Unknown or unidentifiable processor "%s"' % (proc))
 
@@ -626,7 +655,7 @@ def create_template_vars(build_config, options, modules, cc, arch, osinfo):
     def link_to():
         libs = set()
         for module in modules:
-            for (osname,link_to) in module.libs.iteritems():
+            for (osname,link_to) in module.libs.items():
                 if osname == 'all' or osname == osinfo.basename:
                     libs.add(link_to)
                 else:
@@ -722,7 +751,7 @@ def create_template_vars(build_config, options, modules, cc, arch, osinfo):
         'cc': cc.binary_name + cc.mach_abi_link_flags(
             options.os, options.arch, options.cpu, options.debug_build),
 
-        'lib_opt': cc.lib_opt_flags,
+        'lib_opt': cc.library_opt_flags(options),
         'mach_opt': cc.mach_opts(options.arch, options.cpu),
         'check_opt': cc.check_opt_flags,
         'lang_flags': cc.lang_flags + options.extra_flags,
@@ -782,8 +811,7 @@ def create_template_vars(build_config, options, modules, cc, arch, osinfo):
 
         'doc_files': makefile_list(build_config.doc_files()),
 
-        'mod_list': '\n'.join(['%s (%s)' % (m.basename, m.realname)
-                               for m in sorted(modules)]),
+        'mod_list': '\n'.join(sorted([m.basename for m in modules])),
         }
 
 """
@@ -798,7 +826,7 @@ def choose_modules_to_use(options, modules):
     def cannot_use_because(mod, reason):
         not_using_because.setdefault(reason, []).append(mod)
 
-    for (modname, module) in modules.iteritems():
+    for (modname, module) in modules.items():
         if modname in options.disabled_modules:
             cannot_use_because(modname, 'disabled by user')
         elif modname in options.enabled_modules:
@@ -936,16 +964,35 @@ Perform the filesystem operations needed to setup the build
 """
 def setup_build(build_config, options, template_vars):
 
+    """
+    Choose the link method based on system availablity and user request
+    """
+    def choose_link_method(req_method):
+
+        def useable_methods():
+            if 'symlink' in os.__dict__:
+                yield 'symlink'
+            if 'link' in os.__dict__:
+                yield 'hardlink'
+            yield 'copy'
+
+        for method in useable_methods():
+            if req_method is None or req_method == method:
+                return method
+
+        logging.info('Could not use requested link method %s' % (req_method))
+        return 'copy'
+
     """
     Copy or link the file, depending on what the platform offers
     """
-    def portable_symlink(filename, target_dir):
+    def portable_symlink(filename, target_dir, method):
 
         if not os.access(filename, os.R_OK):
             logging.warning('Missing file %s' % (filename))
             return
 
-        if 'symlink' in os.__dict__:
+        if method == 'symlink':
             def count_dirs(dir, accum = 0):
                 if dir == '' or dir == os.path.curdir:
                     return accum
@@ -961,13 +1008,16 @@ def setup_build(build_config, options, template_vars):
 
             os.symlink(source, target)
 
-        elif 'link' in os.__dict__:
+        elif method == 'hardlink':
             os.link(filename,
                     os.path.join(target_dir, os.path.basename(filename)))
 
-        else:
+        elif method == 'copy':
             shutil.copy(filename, target_dir)
 
+        else:
+            raise Exception('Unknown link method %s' % (method))
+
     def choose_makefile_template(style):
         if style == 'nmake':
             return 'nmake.in'
@@ -1027,8 +1077,13 @@ def setup_build(build_config, options, template_vars):
     logging.debug('Linking %d header files in %s' % (
         len(build_config.headers), build_config.full_include_dir))
 
+    link_method = choose_link_method(options.link_method)
+    logging.info('Using %s to link files into build directory' % (link_method))
+
     for header_file in build_config.headers:
-        portable_symlink(header_file, build_config.full_include_dir)
+        portable_symlink(header_file,
+                         build_config.full_include_dir,
+                         link_method)
 
 """
 Main driver
@@ -1059,6 +1114,10 @@ def main(argv = None):
     if options.os == "java":
         raise Exception("Jython detected: need --os and --cpu to set target")
 
+    if re.match('^cygwin_.*', options.os):
+        logging.debug("Converting '%s' to 'cygwin'", options.os)
+        options.os = 'cygwin'
+
     options.base_dir = os.path.dirname(argv[0])
     options.src_dir = os.path.join(options.base_dir, 'src')
 
@@ -1119,9 +1178,9 @@ def main(argv = None):
                 matching_version = '(4\.[01234]\.)|(3\.[34]\.)|(2\.95\.[0-4])'
 
                 gcc_version = ''.join(
-                    subprocess.Popen(['g++', '-v'],
+                    str(subprocess.Popen(['g++', '-v'],
                                      stdout=subprocess.PIPE,
-                                     stderr=subprocess.PIPE).communicate())
+                                     stderr=subprocess.PIPE).communicate()))
 
                 if re.search(matching_version, gcc_version):
                     options.dumb_gcc = True
@@ -1160,7 +1219,7 @@ if __name__ == '__main__':
     try:
         main()
     except Exception, e:
-        print >>sys.stderr, e
+        logging.error(str(e))
         #import traceback
         #traceback.print_exc(file=sys.stderr)
         sys.exit(1)
